/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

package tiledownloader;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.nio.file.Files;
import java.security.GeneralSecurityException;
import java.security.SecureRandom;
import java.security.cert.X509Certificate;
import javax.net.ssl.HttpsURLConnection;
import javax.net.ssl.SSLContext;
import javax.net.ssl.TrustManager;
import javax.net.ssl.X509TrustManager;

/**
 *
 * @author Jorgen
 */
public class DownloadThread implements Runnable
{
    public static final int             MAXRETRIES          =15;
    public static final int             MAXBACKOFFDELAY     =60000;
    public static final int             STARTBACKOFFDELAY   =1000;
    
    private final String                server;
    private final Thread                thread;
    private final int                   id;
    private final int                   zoom;
    private final int                   delay;
    private final File                  defaultImageFile;
    private final FinishedListener      listener;

    private boolean                     threadExit;
    private boolean                     threadRunning;
    private String                      directory;
    private boolean                     downloadingImage;
    private DownloadRequest             request;
    
    // Provision to accept every HTTPS certificate
    static 
    {
        final TrustManager[] trustAllCertificates = new TrustManager[] 
        {
            new X509TrustManager() {
                @Override
                public X509Certificate[] getAcceptedIssuers() {
                    return null; // Not relevant.
                }
                @Override
                public void checkClientTrusted(X509Certificate[] certs, String authType) {
                    // Do nothing. Just allow them all.
                }
                @Override
                public void checkServerTrusted(X509Certificate[] certs, String authType) {
                    // Do nothing. Just allow them all.
                }
            }
        };

        try 
        {
            SSLContext sc = SSLContext.getInstance("SSL");
            sc.init(null, trustAllCertificates, new SecureRandom());
            HttpsURLConnection.setDefaultSSLSocketFactory(sc.getSocketFactory());
        } 
        catch (GeneralSecurityException e) 
        {
            throw new ExceptionInInitializerError(e);
        }
    }    
    
    
    /**
     * Gets the id
     * @return The id
     */
    public int getId()
    {
        return id;
    }
    
    /**
     * Returns the download server
     * @return The tile server
     */
    public String getServer()
    {
        return this.server;
    }
    
    
    /**
     * Constructor. Sets the parameters, starts the thread
     * @param server Server URL
     * @param id Id of the thread
     * @param directory Directory to save the downloaded images to
     * @param defaultImage Image to use if the image cannot be downloaded, or "" if no image to use
     * @param zoom Zoom level
     * @param listener Finished listener, called when download finished
     */
    public DownloadThread(String server, int id, String directory, String defaultImage, int zoom, int delay, FinishedListener listener)
    {
        
        this.server     =server;
        this.id         =id;
        this.directory  =directory;
        this.delay      =delay;
        this.listener   =listener;
        
        
        thread=new Thread(this);
        
        synchronized(this)
        {
            this.zoom               =zoom;
            this.directory          =directory;
            if (!defaultImage.equals(""))
            {
                this.defaultImageFile   =new File(defaultImage);
            }
            else
            {
                this.defaultImageFile   =null;
            }
            threadRunning           =false;
            threadExit              =false;
            this.downloadingImage   =false;
        }
        thread.start();
    }
    
    
    /**
     * Download the tile 
     * @param zoom
     * @param x
     * @param y
     * @param targetDirectory
     * @returns 0 if succeeded, 1 if error
     */
    private boolean downloadImage(int zoom, int x, int y, String targetDirectory, String server)
//            throws MalformedURLException, IOException, FileNotFoundException
    {
        String          url;
        String          filename;
        String          dir;
        String          fullFileName=null;
        File            dirFile;
        File            outputFile;
        boolean         ready;
        int             retries;
        int             backOffDelay;
        int             maxBackOffDelay;
        int             maxRetries;
        boolean         error;
        InputStream     imageReader=null;
        OutputStream    imageWriter=null;

        
        error   = false;
        
        // Initialize
        synchronized(this)
        {
            backOffDelay    =this.STARTBACKOFFDELAY;
            maxRetries      =this.MAXRETRIES; 
            maxBackOffDelay = this.MAXBACKOFFDELAY;
        }
                
        // The directory where to store the png, like 'd:/test/17/12345/'
        dir=targetDirectory + File.separator+zoom+File.separator+x+File.separator;
        dirFile=new File(dir);

        // Create the dir
        if (!dirFile.exists())
        {
            dirFile.mkdirs();
        }

        // The filename
        filename=""+y+".png";
        fullFileName=dir+filename;


        // The URL
        if (server.matches("(\\S*)\\$(\\S*)"))
        {
            url=server;
            url=url.replace("$x", Integer.toString(x));
            url=url.replace("$y", Integer.toString(y));
            url=url.replace("$z", Integer.toString(zoom));
        }
        else
        {
            url=server+"/"+zoom+"/"+x+"/"+y+".png";
        }

        DebugLogger.info("URL: "+url+", Writing to "+ fullFileName);

       
        ready       =false;
        retries     =0;
        
        outputFile=new File(fullFileName);
        
        // If the tile already exists, don't download it again...
        if (!outputFile.exists())
        {
            // The retry loop...
            while (!ready)
            {
                try
                {
                    // Do the download from the URL to file
                    URL imageUrl = new URL(url);

                    int readByte;
                    imageReader = new BufferedInputStream(imageUrl.openStream());
                    imageWriter = new BufferedOutputStream(new FileOutputStream(fullFileName));
                    while ((readByte = imageReader.read()) != -1)
                    {
                        imageWriter.write(readByte);
                    }
                    ready=true;
                }
                catch(MalformedURLException e)
                {
                    // URL is not Ok
                    DebugLogger.error("Malformed URL: "+e.getMessage());
                    ready=true;
                }
                catch(FileNotFoundException e)
                {
                    // If the tile is not found (it often happens for sea tiles,
                    // since these might not be present at the tile server in order
                    // to save space or something), the default tile is copied.
                    // At least, if it exists...
                    if ((defaultImageFile!=null)  && (defaultImageFile.exists()) && !defaultImageFile.isDirectory())
                    {
                        DebugLogger.info("File not found: "+e.getMessage()+" - Copying default image to "+fullFileName);

                        try
                        {
                            Files.copy(defaultImageFile.toPath(), new File(fullFileName).toPath());
                        }
                        catch(IOException ioException)
                        {
                            DebugLogger.error("Error copying file: "+defaultImageFile.toString()+": "+ioException.getMessage());
                        }
                    }
                    else
                    {
                        DebugLogger.info("File not found: "+e.getMessage()); 
                    }
                    ready=true;
                }
                catch (IOException e)
                {
                    // If somehow the request fails (often OpenTopoMap returns 503 - 
                    // service unavailable), retry
                    if (retries<maxRetries)
                    {
                        DebugLogger.info("IO Error: "+e.getMessage()+". Retry: "+(retries+1));
                        // Wait a while
                        try
                        {
                            Thread.sleep(backOffDelay);
                        }
                        catch (InterruptedException sleepException)
                        {
                            DebugLogger.error("Thread sleep interrupted");
                        }

                        // Increase the delay to wait.
                        backOffDelay=Math.min(2*backOffDelay, maxBackOffDelay);
                        retries++;
                    }
                    else
                    {
                        DebugLogger.error("IO Error: image cannot be downloaded in "+maxRetries+" retries");
                        error=true;
                        ready=true;
                    }
                }
                finally
                {
                    try
                    {
                        if (imageReader!=null)
                        {
                            imageReader.close();
                        }
                        if (imageWriter!=null)
                        {
                            imageWriter.close();
                        }
                    }
                    catch (IOException e)
                    {
                        
                    }
                }
            }
        }
        else
        {
            DebugLogger.info("File exists. Skipping download");
        }
        return error;      
    }    
    
    /**
     * Stops the thread.
     */
    public void stop()
    {
        boolean localThreadRunning;
        
        synchronized(this)
        {
            localThreadRunning=threadRunning;
            threadExit=true;
        }
        
        while (localThreadRunning)
        {
            synchronized(this)
            {
                localThreadRunning=threadRunning;
            }
        }
        
    }

    // Signal the thread to stop, don't wait
    public void signalStop()
    {
        synchronized(this)
        {
            threadExit=true;
        }
    }
    
    /**
     * Download the tile at given x,y
     * @param x Index x
     * @param y Index y
     */
    public void doDownloadTile(DownloadRequest request)
    {
        synchronized(this)
        {
            this.request            =request;
            this.downloadingImage   =true;
        }
    }
    
    /**
     * The downloading thread. Waits for a new tile to download and 
     * downloads it
     */
    @Override
    public void run()
    {
        boolean         error;
        boolean         localThreadExit;
        boolean         localDownloadingImage;
        String          localDirectory          =null;
        String          localServer             =null;
        int             localZoom               =0;
        int             localDelay              =100;
        DownloadRequest localRequest    =null;
        
        synchronized(this)
        {
            localThreadExit=threadExit;
            threadRunning=true;
        }
        
        DebugLogger.info("Thread "+this.id+" started");
        
        while (!localThreadExit)
        {
            synchronized(this)
            {
                localDownloadingImage=this.downloadingImage;
                if (localDownloadingImage)
                {
                    localZoom       =this.zoom;

                    localDirectory  =this.directory;
                    localServer     =this.server;
                    localDelay      =this.delay;
                    localRequest    =this.request;
                }
            }
            
            if (localDownloadingImage)
            {
                error=this.downloadImage(localZoom, localRequest.x, localRequest.y, localDirectory, localServer);
                synchronized(this)
                {
                    // Set download finished, each pass
                    this.downloadingImage=false;
                }

                // Signal the main process this thread is ready for the next download
                // or that it failed
                if (!error)
                {
                    this.listener.notifyFinished(id);
                }
                else
                {
                    this.listener.notifyFailed(id, request);
                }
            }
            
            try
            {
                Thread.sleep(localDelay);
            }
            catch(InterruptedException e)
            {
                DebugLogger.error("Thread sleep interrupted: "+e.getMessage());
            }
            
            synchronized(this)
            {
                localThreadExit=threadExit;
            }
        }
        synchronized(this)
        {
            threadRunning=false;
        }
    }
    
    
}
